/*----------------------------------------------------------------------
  File    : jim.c
  Contents: item set mining based on cover similarity
            (especially the generalized Jaccard index)
  Author  : Christian Borgelt
  History : 2010.04.22 file created from eclat.c
            2010.04.23 recursive definition of union exploited
            2010.04.29 bug in memory organization fixed (if n > m)
            2010.07.02 other item cover similarity measures added
            2010.07.04 parameters for generic similarity measure added
            2010.07.12 filter version for intersection/difference added
            2010.07.14 output file made optional (for benchmarking)
            2010.08.19 item selection file added as optional input
            2010.08.22 adapted to modified modules tract and tabread
            2010.10.07 bug in function f_diff fixed (base support)
            2010.10.15 adapted to modified interface of module report
            2010.11.24 adapted to modified error reporting (tract)
            2010.11.26 memory handling simplified (list base sizes etc.)
            2010.12.11 adapted to a generic error reporting function
            2010.12.20 adapted to function tbg_icnts() (filter problem)
            2011.03.20 optional integer transaction weights added
            2011.05.16 bug in function generic_2 fixed (numerator)
            2011.05.17 size-weighted similarity as evaluation measure
            2011.07.08 adapted to modified function tbg_recode()
            2011.08.28 output of item set counters per size added
            2013.03.29 adapted to type changes in module tract
            2013.10.15 checks of return code of isr_report() added
            2013.10.18 optional pattern spectrum collection added
            2013.11.12 item selection file changed to option -R#
            2014.05.12 option -F# added (support border for filtering)
            2014.08.25 adapted to modified item set reporter interface
            2014.10.24 changed from LGPL license to MIT license
            2018.03.21 jim miner object and interface introduced
            2019.04.15 bug in closed/maximal filtering fixed
------------------------------------------------------------------------
  Reference for the JIM algorithm:
    Item Set Mining Based on Cover Similarity.
    M. Segond and C. Borgelt.
    Proc. 15th Pacific-Asia Conf. on Knowledge Discovery and Data Mining
    (PAKDD 2011, Shenzhen, China), LNCS 6635:493--505.
    Springer-Verlag, Berlin, Germany 2011
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifndef ISR_PATSPEC
#define ISR_PATSPEC
#endif
#ifndef ISR_CLOMAX
#define ISR_CLOMAX
#endif
#ifdef  JIM_MAIN
#ifndef PSP_REPORT
#define PSP_REPORT
#endif
#ifndef TA_READ
#define TA_READ
#endif
#endif
#ifdef JIM_ABORT
#include "sigint.h"
#endif
#include "jim.h"
#ifdef JIM_MAIN
#include "params.h"
#include "error.h"
#endif
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "jim"
#define DESCRIPTION "Jaccard item set mining / " \
                    "item set mining based on cover similarity"
#define VERSION     "version 3.19 (2019.04.15)        " \
                    "(c) 2010-2019   Christian Borgelt"

/* --- error codes --- */
/* error codes   0 to  -4 defined in tract.h */
#define E_STDIN      (-5)       /* double assignment of stdin */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* too few/many arguments */
#define E_TARGET     (-9)       /* invalid target type */
#define E_SIZE      (-10)       /* invalid item set size */
#define E_SUPPORT   (-11)       /* invalid item set support */
#define E_SIMILTY   (-12)       /* invalid similarity measure */
#define E_MEASURE   (-13)       /* invalid evaluation measure */
/* error codes -15 to -25 defined in tract.h */

#define DIFFSIZE(p,q) ((size_t)((int*)(p)-(int*)(q)) *sizeof(int))

#ifndef QUIET                   /* if not quiet version, */
#define MSG         fprintf     /* print messages */
#define XMSG        if (jim->mode & JIM_VERBOSE) fprintf
#define CLOCK(t)    ((t) = clock())
#else                           /* if quiet version, */
#define MSG(...)    ((void)0)   /* suppress messages */
#define XMSG(...)   ((void)0)
#define CLOCK(t)    ((void)0)
#endif

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef double SIMFN (SUPP s, SUPP r, SUPP n, double *p);

typedef struct {                /* --- trans. identifier list --- */
  ITEM     item;                /* item / last item in the set */
  SUPP     all;                 /* support (cover   size, all items) */
  SUPP     one;                 /* extent  (carrier size, one item) */
  TID      off;                 /* offset to second tid list */
  double   sim;                 /* item cover similarity measure */
  TID      tids[2];             /* transaction identifiers */
} TIDLIST;                      /* (transaction identifier list) */

struct _jim {                   /* --- Jaccard item set miner --- */
  int      target;              /* target type (e.g. closed/maximal) */
  double   smin;                /* minimum support of an item set */
  SUPP     supp;                /* minimum support of an item set */
  double   wmin;                /* minimum transaction weight */
  ITEM     zmin;                /* minimum size of a rule/item set */
  ITEM     zmax;                /* maximum size of a rule/item set */
  int      eval;                /* additional evaluation measure */
  double   thresh;              /* threshold for evaluation measure */
  int      covsim;              /* cover similarity measure */
  SIMFN    *simfn;              /* similarity measure function */
  double   simps[8];            /* cover similarity parameters */
  double   sim;                 /* cover similarity threshold */
  int      algo;                /* variant of jim algorithm */
  int      mode;                /* search mode (e.g. pruning) */
  SUPP     ttw;                 /* total transaction weight */
  SUPP     *muls;               /* multiplicities of transactions */
  SUPP     *all;                /* flags for all items in prefix */
  SUPP     *none;               /* flags for no  item  in prefix */
  TABAG    *tabag;              /* transaction bag/multiset */
  ISREPORT *report;             /* item set reporter */
};                              /* (split and merge miner) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#if !defined QUIET && defined JIM_MAIN
/* --- error messages --- */
static const char *errmsgs[] = {
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_TARGET   -9 */  "invalid target type '%c'",
  /* E_SIZE    -10 */  "invalid item set size %"ITEM_FMT,
  /* E_SUPPORT -11 */  "invalid minimum support %g",
  /* E_SIMILTY -12 */  "invalid similarity measure '%c'",
  /* E_MEASURE -13 */  "invalid evaluation measure '%c'",
  /*           -14 */  NULL,
  /* E_NOITEMS -15 */  "no (frequent) items found",
  /*           -16 */  "unknown error"
};
#endif

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
#ifdef JIM_MAIN
#ifndef QUIET
static CCHAR    *prgname;       /* program name for error messages */
#endif
static TABREAD  *tread  = NULL; /* table/transaction reader */
static ITEMBASE *ibase  = NULL; /* item base */
static TABAG    *tabag  = NULL; /* transaction bag/multiset */
static ISREPORT *report = NULL; /* item set reporter */
static TABWRITE *twrite = NULL; /* table writer for pattern spectrum */
static double   *border = NULL; /* support border for filtering */
static JIM      *jim    = NULL; /* jim miner object */
#endif

/*----------------------------------------------------------------------
  Similarity Measure Functions
----------------------------------------------------------------------*/

static double constant        (SUPP s, SUPP r, SUPP n, double *p)
{ return 0; }

static double russel_rao      (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)s/(double)n; }
/* S_R = s/n */

static double kulczynski      (SUPP s, SUPP r, SUPP n, double *p)
{ SUPP q = r-s; return (q > 0) ? (double)s/(double)q : INFINITY; }
/* S_K = s/q = s/(r-s) */

static double jaccard         (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)s/(double)r; }
/* S_J = s/r */

static double dice            (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)(s+s)/(double)(r+s); }
/* S_D = 2s/(r+s) */

static double sokal_sneath_1  (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)s/(double)(r+r-s); }
/* S_S = s/(r+q) = s/(2r-s) */

static double sokal_michener  (SUPP s, SUPP r, SUPP n, double *p)
{ return (double)(s+(n-r))/(double)n; }
/* S_M = (s+z)/n */

static double faith           (SUPP s, SUPP r, SUPP n, double *p)
{ return ((double)s+0.5*(double)(n-r))/(double)n; }
/* S_F = (s+z/2)/n */

static double rogers_tanimoto (SUPP s, SUPP r, SUPP n, double *p)
{ SUPP q = r-s; return (double)(n-q)/(double)(n+q); }
/* S_T = (s+z)/(n+q) = (n-q)/(n+q) */

static double sokal_sneath_2  (SUPP s, SUPP r, SUPP n, double *p)
{ SUPP sz = s+(n-r); return (double)(sz+sz)/(double)(n+sz); }
/* S_N = 2(s+z)/(n+s+z) */

static double sokal_sneath_3  (SUPP s, SUPP r, SUPP n, double *p)
{ SUPP q = r-s; return (q > 0) ? (double)(n-q)/(double)q : INFINITY; }
/* S_O = (s+z)/q = (m-q)/q */

static double baroni_buser    (SUPP s, SUPP r, SUPP n, double *p)
{ double x = sqrt((double)s*(n-r)); return (x+(double)s)/(x+(double)r);}
/* S_B = (x+s)/(x+r) */

/*--------------------------------------------------------------------*/

static double generic_0       (SUPP s, SUPP r, SUPP n, double *p)
{                               /* --- generic similarity measure */
  double x = sqrt((double)s*(double)r);
  double d = p[4]*(double)s +p[5]*(double)r +p[6]*(double)n +p[7]*x;
  if (fabs(d) <= DBL_MIN) return INFINITY;
  return (p[0]*s +p[1]*r +p[2]*n +p[3]*x) /d;
}  /* generic_0() */
/* S = (c_0s +c_1z +c_2n +c_3x) / (c_4s +c_5z +c_6n +c_7x) */

/*--------------------------------------------------------------------*/

static double generic_1       (SUPP s, SUPP r, SUPP n, double *p)
{                               /* --- generic similarity measure */
  double d = p[4]*(double)s +p[5]*(double)r +p[6]*(double)n;
  if (fabs(d) <= DBL_MIN) return INFINITY;
  return (p[0]*(double)s +p[1]*(double)r +p[2]*(double)n) /d;
}  /* generic_1() */
/* S = (c_0s +c_1z +c_2n) / (c_4s +c_5z +c_6n) */

/*--------------------------------------------------------------------*/

static double generic_2       (SUPP s, SUPP r, SUPP n, double *p)
{                               /* --- generic similarity measure */
  double d = p[4]*(double)s +p[5]*(double)r +p[6]*(double)n;
  return (fabs(d) > DBL_MIN)
       ? (p[0]*(double)s +p[1]*(double)r) /d : INFINITY;
}  /* generic_2() */
/* S = (c_0s +c_1z) / (c_4s +c_5z +c_6n) */

/*--------------------------------------------------------------------*/

static double generic_3       (SUPP s, SUPP r, SUPP n, double *p)
{                               /* --- generic similarity measure */
  double d = p[4]*(double)s +p[5]*(double)r +p[6]*(double)n;
  return (fabs(d) > DBL_MIN) ? p[0]*(double)s /d : INFINITY;
}  /* generic_3() */
/* S = c_0s / (c_4s +c_5z +c_6n) */

/*--------------------------------------------------------------------*/

static SIMFN *simfns[] = {
  /* JIM_NONE             0 */  constant,
  /* JIM_RUSSEL_RAO       1 */  russel_rao,
  /* JIM_KULCZYNSKI       2 */  kulczynski,
  /* JIM_JACCARD          3 */  jaccard,
  /* JIM_DICE             4 */  dice,
  /* JIM_SOKAL_SNEATH_1   5 */  sokal_sneath_1,
  /* JIM_SOKAL_MICHENER   6 */  sokal_michener,
  /* JIM_FAITH            7 */  faith,
  /* JIM_ROGERS_TANIMOTO  8 */  rogers_tanimoto,
  /* JIM_SOKAL_SNEATH_2   9 */  sokal_sneath_2,
  /* JIM_SOKAL_SNEATH_3  10 */  sokal_sneath_3,
  /* JIM_BARONI_BUSER    11 */  baroni_buser,
  /* JIM_GENERIC         12 */  generic_0,
};

/*----------------------------------------------------------------------
  Auxiliary Functions (for debugging)
----------------------------------------------------------------------*/
#ifndef NDEBUG

static void indent (int k)
{ while (--k >= 0) printf("   "); }

/*--------------------------------------------------------------------*/

static void show (const char *text, TIDLIST **lists, ITEM k, int ind)
{                               /* --- show a cond. trans. database */
  ITEM i;                       /* loop variable */
  TID  *t;                      /* to traverse transaction ids */

  indent(ind);                  /* indent the output line */
  printf("%s\n", text);         /* print the given text */
  for (i = 0; i < k; i++) {     /* traverse the item lists */
    indent(ind);                /* indent the output line */
    printf("%s:", ib_name(ibase, lists[i]->item));
    for (t = lists[i]->tids; *t >= 0; t++)
      printf(" %"TID_FMT, *t);  /* print the item and the tid list */
    printf(" :");               /* print separator to diff. list */
    for (t++; *t >= 0; t++)     /* skip the sentinel marker and */
      printf(" %"TID_FMT, *t);  /* print list of transaction ids */
    printf("\n");               /* finally terminate the line */
  }
}  /* show() */

#endif  /* #ifndef NDEBUG */
/*----------------------------------------------------------------------
  Jaccard Item Set Mining
----------------------------------------------------------------------*/

static SUPP isect (TIDLIST *dst, TIDLIST *src1,TIDLIST *src2,SUPP *muls)
{                               /* --- intersect two tid lists */
  TID *s1, *s2, *d;             /* to traverse sources and dest. */

  assert(dst && src1 && src2 && muls); /* check function arguments */
  dst->item = src1->item;       /* copy the first item and */
  dst->all  = 0;                /* initialize the support value */
  s1 = src1->tids; s2 = src2->tids; d = dst->tids;
  while (1) {                   /* tid list intersection loop */
    if      (*s1 > *s2) s1++;   /* if one tid is larger, */
    else if (*s1 < *s2) s2++;   /* simply skip this tid */
    else if (*s1 < 0) break;    /* check for the sentinel */
    else { dst->all += muls[*d++ = *s1++]; s2++; }
  }                             /* copy equal elements to destination */
  *d++ = (TID)-1;               /* store a sentinel at the list end */
  dst->off = (TID)(d-dst->tids);/* note offset to the second tid list */
  return dst->all;              /* return the item (set) support */
}  /* isect() */

/*--------------------------------------------------------------------*/

static TID diff (TIDLIST *dst, TIDLIST *src1, TIDLIST *src2,SUPP *muls)
{                               /* --- subtract two tid lists */
  TID *s1, *s2, *d;             /* to traverse sources and dest. */

  assert(dst && src1 && src2 && muls); /* check function arguments */
  dst->one = src2->one;         /* initialize the support value */
  s1 = src1->tids +src1->off;   /* get the tid lists */
  s2 = src2->tids +src2->off; d = dst->tids +dst->off;
  while (1) {                   /* tid list difference loop */
    if      (*s1 > *s2) dst->one += muls[*d++ = *s1++];
    else if (*s1 < *s2) s2++;   /* copy all tids to the destination */
    else if (*s1 < 0) break;    /* that occur only in the first list */
    else { s1++; s2++; }        /* (difference set for union */
  }                             /* of covers of single items) */
  *d++ = (TID)-1;               /* store a sentinel at the list end */
  return (TID)(d -dst->tids);   /* return the size of the new lists */
}  /* diff() */

/*----------------------------------------------------------------------
It is advantageous to separate the intersection and the difference set,
because the difference set computation need not be carried out if the
intersection yields a support below the minimum support. However, the
disadvantage of the separation is that one needs an offset field in
the transaction id lists to point to the start of the second tid list.
----------------------------------------------------------------------*/

static SUPP f_isect (TIDLIST *dst, TIDLIST *src, SUPP *muls)
{                               /* --- filter a tid list (i.section) */
  SUPP f;                       /* transaction frequency */
  TID  *s, *d;                  /* to traverse source and dest. */

  assert(dst && src && muls);   /* check function arguments */
  dst->item = src->item;        /* copy the first item and */
  dst->all  = 0;                /* initialize the support */
  d = dst->tids;                /* traverse the source tid list */
  for (s = src->tids; *s >= 0; s++) {
    f = muls[*s];               /* collect the marked tids */
    if (f > 0) { dst->all += f; *d++ = *s; }
  }                             /* sum the transaction weights */
  *d++ = (TID)-1;               /* store a sentinel at the list end */
  dst->off = (TID)(d-dst->tids);/* note offset to the second tid list */
  return dst->all;              /* return the time set support */
}  /* f_isect() */

/*--------------------------------------------------------------------*/

static TID f_diff (TIDLIST *dst, TIDLIST *src, SUPP *muls, SUPP base)
{                               /* --- filter a tid list (difference) */
  SUPP m;                       /* transaction multiplicity */
  TID  *s, *d;                  /* to traverse source and dest. */

  assert(dst && src && muls);   /* check function arguments */
  dst->item = src->item;        /* copy the first item and */
  dst->one  = base;             /* initialize the support */
  d = dst->tids +dst->off;      /* traverse the tid list */
  for (s = src->tids +src->off; *s >= 0; s++) {
    m = muls[*s];               /* collect the marked tids */
    if (m > 0) { dst->one += m; *d++ = *s; }
  }                             /* sum the transaction weights */
  *d++ = (TID)-1;               /* store a sentinel at the list end */
  return (TID)(d -dst->tids);   /* return the size of the new lists */
}  /* f_diff() */

/*----------------------------------------------------------------------
Since multiple intersections/differences have to be computed with the
same tid list, it pays to convert the tid list into a flag array,
so that no tid comparisons are required to compute the results.
----------------------------------------------------------------------*/

static int recurse (JIM *jim, TIDLIST **lists, ITEM k, size_t x)
{                               /* --- jim recursion */
  int     r;                    /* error status */
  ITEM    i, m;                 /* loop variables */
  SUPP    s;                    /* support buffer */
  SUPP    pex;                  /* minimum support for perfect exts. */
  TIDLIST **proj = NULL;        /* tid lists of projected database */
  TIDLIST *l, *d;               /* to traverse the tid lists */
  TID     *p;                   /* to traverse the tids */
  double  j;                    /* similarity measure of item set */

  assert(jim && lists && (k > 0)); /* check the function arguments */
  #ifdef JIM_ABORT              /* if to check for interrupt */
  if (sig_aborted()) return -1; /* if execution was aborted, */
  #endif                        /* abort the recursion */
  if ((k > 1)                   /* if there is more than one item */
  &&  isr_xable(jim->report,2)){/* and another item can be added */
    proj = (TIDLIST**)malloc((size_t)k *sizeof(TIDLIST*) +x);
    if (!proj) return -1;       /* allocate list and element arrays */
  }                             /* (memory for projected database) */
  for (r = 0; --k >= 0; ) {     /* traverse the items/tid lists */
    l = lists[k];               /* get the current tid list */
    r = (jim->mode & JIM_AP4CM) /* if aggressive closed/max. pruning */
      ? isr_addwgt(jim->report, l->item, l->all, l->sim)
      : isr_addncw(jim->report, l->item, l->all, l->sim);
    if (r <  0) break;          /* add current item to the reporter */
    if (r <= 0) continue;       /* check if item needs processing */
    if (proj) {                 /* if another item can be added */
      pex = (jim->mode & JIM_PERFECT) ? l->all : ITEM_MAX;
      proj[m = 0] = d = (TIDLIST*)(proj +k+1);
      if (k < 2) {              /* if there are only few items left */
        /* Benchmark tests showed that this version is faster only */
        /* if there is only one other tid list to intersect with.  */
        for (i = 0; i < k; i++) {
          s = isect(d, lists[i], l, jim->muls);
          if (s < jim->supp)    /* intersect with preceding tid lists */
            continue;           /* and check against mininum support */
          x = (size_t)diff(d, lists[i], l, jim->muls);
          j = jim->simfn(s, d->one, jim->ttw, jim->simps);
          if (j < jim->sim)     /* compute the similarity measure and */
            continue;           /* check it against given minimum */
          if ((s >= pex)        /* collect perfect extensions */
          &&  (j >= isr_wgt(jim->report))) {
            isr_addpex(jim->report, d->item); continue; }
          d->sim = j;           /* record cover similarity measure */
          proj[++m] = d = (TIDLIST*)(d->tids +x);
        } }                     /* collect the qualifying tid lists */
      else {                    /* if there are many items left */
        p = l->tids-1;          /* mark/unmark tids in current lists */
        while (*++p >= 0) jim->all [*p] = jim->muls[*p];
        while (*++p >= 0) jim->none[*p] = 0;
        for (i = 0; i < k; i++) {
          s = f_isect(d, lists[i], jim->all);
          if (s < jim->supp)    /* intersect with preceding tid lists */
            continue;           /* and check against minimum support */
          x = (size_t)f_diff(d, lists[i], jim->none, l->one);
          j = jim->simfn(s, d->one, jim->ttw, jim->simps);
          if (j < jim->sim)     /* compute the similarity measure and */
            continue;           /* check it against given minimum */
          if ((s >= pex)        /* collect perfect extensions */
          &&  (j >= isr_wgt(jim->report))) {
            isr_addpex(jim->report, d->item); continue; }
          d->sim = j;           /* record cover similarity measure */
          proj[++m] = d = (TIDLIST*)(d->tids +x);
        }                       /* collect the qualifying tid lists */
        p = l->tids -1;         /* unmark/mark tids in current lists */
        while (*++p >= 0) jim->all [*p] = 0;
        while (*++p >= 0) jim->none[*p] = jim->muls[*p];
      }
      if (m > 0) {              /* if the projection is not empty */
        r = recurse(jim, proj, m, DIFFSIZE(d, proj[0]));
        if (r < 0) break;       /* recursively find freq. item sets */
      }                         /* in the created projection */
    }
    r = isr_report(jim->report);/* report the current item set */
    if (r < 0) break;           /* and check for an error */
    isr_remove(jim->report, 1); /* remove the current item */
  }                             /* from the item set reporter */
  if (proj) free(proj);         /* delete the list and element arrays */
  return r;                     /* return the error status */
}  /* recurse() */

/*--------------------------------------------------------------------*/

int jim_base (JIM *jim)
{                               /* --- search for frequent item sets */
  int        r = 0;             /* result of recursion/error status */
  ITEM       i, k, m;           /* loop variable, number of items */
  TID        j, n;              /* loop variable, number of trans. */
  size_t     x;                 /* number of item instances */
  SUPP       w;                 /* weight/support buffer */
  TRACT      *t;                /* to traverse the transactions */
  TIDLIST    **lists, *l;       /* to traverse the tid lists */
  TID        *tids, *p, **next; /* to traverse transaction ids */
  const ITEM *s;                /* to traverse transaction items */
  const TID  *c;                /* item occurrence counters */

  assert(jim);                  /* check the function arguments */
  jim->ttw = tbg_wgt(jim->tabag);    /* initialize the recursion data */
  if (jim->ttw < jim->supp)     /* check the total transaction weight */
    return 0;                   /* against the minimum support */
  isr_setwgt(jim->report, 1);   /* set the base similarity value */
  k = tbg_itemcnt(jim->tabag);  /* get the number of items and */
  if (k <= 0) return 0;         /* check for at least one item */
  c = tbg_icnts(jim->tabag, 0); /* get the number of containing */
  if (!c) return -1;            /* transactions per item and */
  n = tbg_cnt(jim->tabag);      /* the number of transactions */
  lists = (TIDLIST**)malloc((size_t) k      *sizeof(TIDLIST*)
                           +(size_t) k      *sizeof(TID*)
                           +(size_t)(n+n+n) *sizeof(SUPP));
  if (!lists) return -1;        /* create initial tid lists array */
  next      = (TID**)(lists+k); /* and split off next position array */
  jim->muls = (SUPP*)(next +k); /* transaction multiplicity array */
  jim->none = jim->muls +n;     /* and the flags/weights arrays */
  jim->all  = jim->none +n;     /* for the filter versions */
  memset(jim->all, 0, (size_t)n *sizeof(SUPP));
  x = tbg_extent(jim->tabag);   /* get the number of item occurrences */
  p = tids = (TID*)malloc((size_t)k*sizeof(TIDLIST) +(x+x)*sizeof(TID));
  if (!p) { free(lists); return -1; }  /* allocate tid list elements */
  for (i = 0; i < k; i++) {     /* traverse the items / tid lists */
    lists[i] = l = (TIDLIST*)p; /* get/create the next tid list */
    l->item  = i;               /* initialize the list item */
    l->all   = 0;               /* init. the tid counter and */
    l->off   = (j = c[i]) +1;   /* get the number of tids */
    next[i]  = l->tids;         /* note position of next trans. id */
    p  = l->tids +j; *p++ = (TID)-1;  /* skip the space for */
    p +=          j; *p++ = (TID)-1;  /* the transaction ids */
  }                             /* store a sentinel at the end */
  for (j = n; j > 0; ) {        /* traverse the transactions */
    t = tbg_tract(jim->tabag, --j);   /* get the next transaction */
    jim->muls[j] = w = ta_wgt(t);     /* and store its multiplicity */
    jim->none[j] = w;                 /* set the no item flags */
    for (s = ta_items(t); *s >= 0; s++) {
      l = lists[*s];            /* traverse the transaction's items */
      l->all += w;              /* sum the transaction weight */
      next[*s][l->off] = (TID)j;/* collect the transaction ids */
      *next[*s]++      = (TID)j;/* in both transaction id lists */
    }                           /* (for all and one contained item) */
  }
  for (i = m = 0; i < k; i++) { /* traverse the items/tid lists, */
    l = lists[i];               /* but skip infrequent items */
    if (l->all < jim->supp)  continue;
    l->sim = jim->simfn(l->all, l->one = l->all, jim->ttw, jim->simps);
    if (l->sim < jim->sim) continue;
    lists[m++] = l;             /* compute and check the similarity */
  }                             /* collect lists for frequent items */
  if (m > 0)                    /* find frequent item sets */
    r = recurse(jim, lists, m, DIFFSIZE(p, tids));
  free(tids); free(lists);      /* delete the allocated arrays */
  return r;                     /* return the error status */
}  /* jim_base() */

/*----------------------------------------------------------------------
  Jaccard Item Set Mining (generic)
----------------------------------------------------------------------*/

JIM* jim_create (int target, double smin, ITEM zmin, ITEM zmax,
                 int eval, double thresh,
                 int covsim, double *simps, double sim,
                 int algo, int mode)
{                               /* create a Jaccard item set miner */
  JIM *jim;                     /* created Jaccard item set miner */

  /* --- make parameters consistent --- */
  if      (target & JIM_MAXIMAL) target = ISR_MAXIMAL;
  else if (target & JIM_CLOSED)  target = ISR_CLOSED;
  else                           target = ISR_FREQUENT;
  if (algo == JIM_AUTO)         /* ensure unique target type */
    algo = JIM_BASIC;           /* and choose the algorithm */

  /* --- create a split and merge miner --- */
  jim = (JIM*)malloc(sizeof(JIM));
  if (!jim) return NULL;        /* create a split and merge miner */
  jim->target = target;         /* and store all parameters */
  jim->smin   = smin;
  jim->supp   = 1;
  jim->zmin   = zmin;
  jim->zmax   = zmax;
  jim->eval   = eval;
  jim->thresh = thresh/100.0;
  jim->covsim = covsim;
  if ((covsim < JIM_GENERIC) || !simps) {
    jim->simfn = simfns[covsim];
    memset(jim->simps, 0, 8*sizeof(double)); }
  else {                        /* set similarity function & params. */
    jim->simps[0] = simps[0];
    jim->simps[1] =         -simps[1];
    jim->simps[2] = simps[2]+simps[1];
    jim->simps[3] = simps[3];
    jim->simps[4] = simps[4];
    jim->simps[5] =         -simps[5];
    jim->simps[6] = simps[6]+simps[5];
    jim->simps[7] = simps[7];
    if      ((fabs(jim->simps[3]) > DBL_MIN)
    ||       (fabs(jim->simps[7]) > DBL_MIN))
         jim->simfn = generic_0;  /* use all parameters */
    else if  (fabs(jim->simps[2]) > DBL_MIN)
         jim->simfn = generic_1;  /* do not use geom. mean of s & z */
    else if  (fabs(jim->simps[1]) > DBL_MIN)
         jim->simfn = generic_2;  /* do not use zero zero counter */
    else jim->simfn = generic_3;  /* use normalized support */
  }
  jim->sim    = sim/100;        /* cover similarity threshold */
  jim->algo   = algo;
  jim->mode   = mode;
  jim->tabag  = NULL;
  jim->report = NULL;
  return jim;                   /* return created item set miner */
}  /* jim_create() */

/*--------------------------------------------------------------------*/

void jim_delete (JIM *jim, int deldar)
{                               /* --- delete Jaccard item set miner */
  if (deldar) {                 /* if to delete data and reporter */
    if (jim->report) isr_delete(jim->report, 0);
    if (jim->tabag)  tbg_delete(jim->tabag,  1);
  }                             /* delete if existing */
  free(jim);                    /* delete the base structure */
}  /* jim_delete() */

/*--------------------------------------------------------------------*/

int jim_data (JIM *jim, TABAG *tabag, int sort)
{                               /* --- prepare data for JIM */
  ITEM    m;                    /* number of items */
  double  smin;                 /* absolute minimum support */
  SUPP    w;                    /* total transaction weight */
  #ifndef QUIET                 /* if to print messages */
  TID     n;                    /* number of transactions */
  clock_t t;                    /* timer for measurements */
  #endif                        /* (only needed for messages) */

  assert(jim && tabag);         /* check the function arguments */
  jim->tabag = tabag;           /* note the transaction bag */
  w = tbg_wgt(tabag);           /* compute absolute minimum support */
  smin = ceilsupp((jim->smin < 0) ? -jim->smin
                : (jim->smin/100.0) *(double)w *(1-DBL_EPSILON));
  jim->supp = (SUPP)ceilsupp(smin);

  /* --- sort and recode items --- */
  CLOCK(t);                     /* start timer, print log message */
  XMSG(stderr, "filtering, sorting and recoding items ... ");
  m = tbg_recode(tabag, jim->supp, -1, -1, -sort);
  if (m <  0) return E_NOMEM;   /* recode items and transactions */
  if (m <= 0) return E_NOITEMS; /* and check the number of items */
  XMSG(stderr, "[%"ITEM_FMT" item(s)]", m);
  XMSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- sort and reduce transactions --- */
  CLOCK(t);                     /* start timer, print log message */
  XMSG(stderr, "sorting and reducing transactions ... ");
  tbg_itsort(tabag, -1, 0);     /* sort items in transactions and */
  tbg_sort  (tabag, -1, 0);     /* sort the trans. lexicographically */
  tbg_reduce(tabag, 0);         /* reduce transactions to unique ones */
  #ifndef QUIET                 /* if to print messages */
  n = tbg_cnt(tabag);           /* get the number of transactions */
  w = tbg_wgt(tabag);           /* and the new transaction weight */
  XMSG(stderr, "[%"TID_FMT, n); /* print number of transactions */
  if (w != (SUPP)n) { XMSG(stderr, "/%"SUPP_FMT, w); }
  XMSG(stderr, " transaction(s)] done [%.2fs].\n", SEC_SINCE(t));
  #endif
  return 0;                     /* return 'ok' */
}  /* jim_data() */

/*--------------------------------------------------------------------*/

int jim_report (JIM *jim, ISREPORT *report)
{                               /* --- prepare reporter for JIM */
  TID n;                        /* number of transactions */
  int mrep;                     /* mode for item set reporter */

  assert(jim && report);        /* check the function arguments */
  jim->report = report;         /* note the item set reporter */

  /* --- get reporting mode --- */
  mrep = 0;                     /* init. the reporting mode */
  #ifdef USE_ZLIB               /* if optional output compression */
  if (jim->mode & JIM_ZLIB)     /* if the compression flag is set, */
    mrep |= ISR_ZLIB;           /* transfer it to the report mode */
  #endif

  /* --- configure item set reporter --- */
  isr_setsupp(report, (RSUPP)jim->supp, RSUPP_MAX);
  isr_setsize(report, jim->zmin, jim->zmax);
  if      (jim->eval == JIM_LDRATIO) /* set add. evaluation measure */
    isr_seteval(report, isr_logrto,  NULL, +1, jim->thresh);
  else if (jim->eval == JIM_SIZESIM)
    isr_seteval(report, isr_sizewgt, NULL, +1, jim->thresh);
  n = (jim->mode & JIM_PREFMT)  /* get range of numbers to preformat */
    ? (TID)ib_maxfrq(tbg_base(jim->tabag)) : -1;
  if ((isr_prefmt(report, (TID)jim->supp, n)      != 0)
  ||  (isr_settarg(report, jim->target, mrep, -1) != 0))
    return E_NOMEM;             /* set pre-format and target type */
  return 0;                     /* return 'ok' */
}  /* jim_report() */

/*--------------------------------------------------------------------*/

int jim_mine (JIM *jim)
{                               /* --- JIM algorithm (generic) */
  int     r;                    /* result of function call */
  #ifndef QUIET                 /* if to print messages */
  clock_t t;                    /* timer for measurements */
  #endif                        /* (only needed for messages) */

  assert(jim);                  /* check the function arguments */
  CLOCK(t);                     /* start timer, print log message */
  r = jim_base(jim);            /* basic frequent item set search */
  if (r < 0) return E_NOMEM;    /* search for frequent item sets */
  XMSG(stderr, "[%"ISCNT_FMT" set(s)]",
       ISCOUT(isr_repcnt(jim->report)));
  XMSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  return 0;                     /* return 'ok' */
}  /* jim_mine() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/
#ifdef JIM_MAIN

static void help (void)
{                               /* --- print add. option information */
  #ifndef QUIET
  fprintf(stderr, "\n");        /* terminate startup message */
  printf("item cover similarity measures (option -i#)\n");
  printf("  r   Russel & Rao                S_R = s/n\n");
  printf("  k   Kulczynski                  S_K = s/q\n");
  printf("  j   Jaccard/Tanimoto            S_J = s/r\n");
  printf("  d   Dice/Sorensen/Czekanowski   S_D = 2s/(r+s)\n");
  printf("  s   Sokal & Sneath 1            S_S = s/(r+q)\n");
  printf("  m   Sokal & Michener/Hamming    S_M = (s+z)/n\n");
  printf("  f   Faith                       S_F = (s+z/2)/n\n");
  printf("  t   Rogers & Tanimoto           S_T = (s+z)/(n+q)\n");
  printf("  n   Sokal & Sneath 2            S_N = 2(s+z)/(n+s+z)\n");
  printf("  o   Sokal & Sneath 3            S_O = (s+z)/q \n");
  printf("  b   Baroni-Urbani & Buser       "
               "S_B = (sqrt(sz)+s)/(sqrt(sz)+r)\n");
  printf("  g   generic similarity measure  "
               "S   = (c0s +c1z +c2n +c3sqrt(sz))\n");
  printf("                                  "
               "    / (c4s +c5z +c6n +c7sqrt(sz))\n");
  printf("      parameters can be specified "
               "with the option -p#:#:#:#:#:#:#:#\n");
  printf("all measures are defined in the following quantities:\n");
  printf("  n   total number of transactions\n");
  printf("  s   number of transactions "
               "with all items in the item set\n");
  printf("  r   number of transactions "
               "with at least one item in the item set\n");
  printf("  q   number of transactions "
               "with some (but not all) items\n");
  printf("  z   number of transactions "
               "with no items in the item set\n");
  printf("\n");
  printf("additional evaluation measures (option -e#)\n");
  printf("  x   no measure (default)\n");
  printf("  b   binary logarithm of support quotient\n");
  printf("  w   similarity measure weighted with itemset size\n");
  printf("\n");
  printf("information output format characters (option -v#)\n");
  printf("  %%%%  a percent sign\n");
  printf("  %%i  number of items (item set size)\n");
  printf("  %%a  absolute item set support\n");
  printf("  %%s  relative item set support as a fraction\n");
  printf("  %%S  relative item set support as a percentage\n");
  printf("  %%w  item cover similarity as a fraction\n");
  printf("  %%W  item cover similarity as a percentage\n");
  printf("  %%e  additional evaluation measure\n");
  printf("  %%E  additional evaluation measure as a percentage\n");
  printf("  %%Q  total transaction weight (database size)\n");
  printf("All format characters can be preceded by the number\n");
  printf("of significant digits to be printed (at most 32 digits),\n");
  printf("even though this value is ignored for integer numbers.\n");
  #endif                        /* print help information */
  exit(0);                      /* abort the program */
}  /* help() */

/*--------------------------------------------------------------------*/

static ITEM getbdr (char *s, char **end, double **border)
{                               /* --- get the support border */
  ITEM   i, k;                  /* loop variables */
  double *b;                    /* support border */

  assert(s && end && border);   /* check the function arguments */
  for (i = k = 0; s[i]; i++)    /* traverse the string and */
    if (s[i] == ':') k++;       /* count the number separators */
  *border = b = (double*)malloc((size_t)++k *sizeof(double));
  if (!b) return -1;            /* allocate a support border */
  for (i = 0; i < k; i++) {     /* traverse the parameters */
    b[i] = strtod(s, end);      /* get the next parameter and */
    if (*end == s) break;       /* check for an empty parameter */
    s = *end; if (*s++ != ':') break;
  }                             /* check for a colon */
  if (++i < k)                  /* shrink support array if possible */
    *border = (double*)realloc(b, (size_t)i *sizeof(double));
  return i;                     /* return number of support values */
}  /* getbdr() */

/*--------------------------------------------------------------------*/

static int setbdr (ISREPORT *report, SUPP w, ITEM zmin,
                   double **border, ITEM n)
{                               /* --- set the support border */
  double s;                     /* to traverse the support values */

  assert(report                 /* check the function arguments */
  &&    (w > 0) && (zmin >= 0) && border && (*border || (n <= 0)));
  while (--n >= 0) {            /* traverse the support values */
    s = (*border)[n];           /* transform to absolute count */
    s = ceilsupp((s >= 0) ? s/100.0 *(double)w *(1-DBL_EPSILON) : -s);
    if (isr_setbdr(report, n+zmin, (RSUPP)s) < 0) return -1;
  }                             /* set support in item set reporter */
  if (*border) { free(*border); *border = NULL; }
  return 0;                     /* return 'ok' */
}  /* setbdr() */

/*--------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (twrite) twr_delete(twrite, 1); \
  if (report) isr_delete(report, 0); \
  if (tabag)  tbg_delete(tabag,  0); \
  if (tread)  trd_delete(tread,  1); \
  if (ibase)  ib_delete (ibase);     \
  if (border) free(border);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse the options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input  file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *fn_sel  = NULL;      /* name of item selection file */
  CCHAR   *fn_psp  = NULL;      /* name of pattern spectrum file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *hdr     = "";        /* record header  for output */
  CCHAR   *sep     = " ";       /* item separator for output */
  CCHAR   *dflt    = " (%S)";   /* default format for check */
  CCHAR   *info    = dflt;      /* format for information output */
  int     target   = 's';       /* target type (closed/maximal) */
  ITEM    zmin     = 2;         /* minimum size of an item set */
  ITEM    zmax     = ITEM_MAX;  /* maximum size of an item set */
  double  smin     = 1;         /* minimum support of an item set */
  int     covsim   = 'j';       /* item cover similarity measure */
  double  simps[8] = { 1, 0, 0, 0,  0, -1, 1, 0 };
  double  sim      = 10;        /* minimum similarity (in percent) */
  int     eval     = 'x';       /* additional evaluation measure */
  double  thresh   = 10;        /* threshold for evaluation measure */
  int     sort     = -2;        /* flag for item sorting and recoding */
  int     algo     = JIM_BASIC; /* variant of JIM algorithm */
  int     mode     = JIM_DEFAULT;  /* search mode (e.g pruning) */
  int     mtar     = 0;         /* mode for transaction reading */
  int     scan     = 0;         /* flag for scanable item output */
  int     bdrcnt   = 0;         /* number of support values in border */
  int     stats    = 0;         /* flag for item set statistics */
  PATSPEC *psp;                 /* collected pattern spectrum */
  ITEM    m;                    /* number of items */
  TID     n;                    /* number of transactions */
  SUPP    w;                    /* total transaction weight */
  #ifndef QUIET                 /* if not quiet version */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no arguments given */
    printf("usage: %s [options] infile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-t#      target type                              "
                    "(default: %c)\n", target);
    printf("         (s: frequent, c: closed, m: maximal item sets)\n");
    printf("-m#      minimum number of items per item set     "
                    "(default: %"ITEM_FMT")\n", zmin);
    printf("-n#      maximum number of items per item set     "
                    "(default: no limit)\n");
    printf("-s#      minimum support of an item set           "
                    "(default: %g%%)\n", smin);
    printf("         (positive: percentage, "
                     "negative: absolute number)\n");
    printf("-i#      item cover similarity measure to use     "
                    "(default: %c)\n", covsim);
    printf("-p#:#..  parameters for generic cover similarity  "
                    "(default: 1:0:0:0:0:-1:1:0)\n");
    printf("-c#      minimum cover similarity of an item set  "
                    "(default: %g%%)\n", sim);
    printf("-e#      additional evaluation measure            "
                    "(default: none)\n");
    printf("-d#      minimum value of add. evaluation measure "
                    "(default: %g%%)\n", thresh);
    printf("-q#      sort items w.r.t. their frequency        "
                    "(default: %d)\n", sort);
    printf("         (1: ascending, -1: descending, 0: do not sort,\n"
           "          2: ascending, -2: descending w.r.t. "
                    "transaction size sum)\n");
    printf("-x       do not prune with perfect extensions     "
                    "(default: prune)\n");
    printf("-F#:#..  support border for filtering item sets   "
                    "(default: none)\n");
    printf("         (list of minimum support values, "
                    "one per item set size,\n");
    printf("         starting at the minimum size, "
                    "as given with option -m#)\n");
    printf("-y       aggressive closed/maximal pruning        "
                    "(default: none)\n");
    printf("         (faster, but often loses item sets)\n");
    printf("-R#      read an item selection from a file\n");
    printf("-P#      write a pattern spectrum to a file\n");
    printf("-Z       print item set statistics "
                    "(number of item sets per size)\n");
    printf("-g       write output in scanable form "
                    "(quote certain characters)\n");
    printf("-h#      record header  for output                "
                    "(default: \"%s\")\n", hdr);
    printf("-k#      item separator for output                "
                    "(default: \"%s\")\n", sep);
    printf("-v#      output format for item set information   "
                    "(default: \"%s\")\n", info);
    printf("-w       integer transaction weight in last field "
                    "(default: only items)\n");
    printf("-r#      record/transaction separators            "
                    "(default: \"\\n\")\n");
    printf("-f#      field /item        separators            "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                       "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                       "
                    "(default: \"#\")\n");
    printf("-!       print additional option information\n");
    printf("infile   file to read transactions from           "
                    "[required]\n");
    printf("outfile  file to write frequent item sets to      "
                    "[optional]\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */
  #endif  /* #ifndef QUIET */
  /* free option characters: ajlouyz [A-Z]\[CFPRZ] */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate switches */
          case '!': help();                          break;
          case 't': target = (*s) ? *s++ : 's';      break;
          case 'm': zmin   = (ITEM)strtol(s, &s, 0); break;
          case 'n': zmax   = (ITEM)strtol(s, &s, 0); break;
          case 's': smin   =       strtod(s, &s);    break;
          case 'i': covsim = (*s) ? *s++ : 'j';      break;
          case 'p': getdblvec(s, &s, 8, simps);      break;
          case 'c': sim    =       strtod(s, &s);    break;
          case 'e': eval   = (*s) ? *s++ : 0;        break;
          case 'd': thresh =       strtod(s, &s);    break;
          case 'F': bdrcnt = getbdr(s, &s, &border); break;
          case 'y': mode  |= JIM_AP4CM;              break;
          case 'R': optarg = &fn_sel;                break;
          case 'P': optarg = &fn_psp;                break;
          case 'Z': stats  = 1;                      break;
          case 'g': scan   = 1;                      break;
          case 'q': sort   = (int) strtol(s, &s, 0); break;
          case 'x': mode  &= ~JIM_PERFECT;           break;
          case 'h': optarg = &hdr;                   break;
          case 'k': optarg = &sep;                   break;
          case 'v': optarg = &info;                  break;
          case 'w': mtar  |= TA_WEIGHT;              break;
          case 'r': optarg = &recseps;               break;
          case 'f': optarg = &fldseps;               break;
          case 'b': optarg = &blanks;                break;
          case 'C': optarg = &comment;               break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (optarg && *s) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fn_inp = s;      break;
        case  1: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg)       error(E_OPTARG);     /* check option arguments */
  if (k      < 1)   error(E_ARGCNT);     /* and number of arguments */
  if (zmin   < 1)   error(E_SIZE, zmin); /* check the size limits */
  if (zmax   < 0)   error(E_SIZE, zmax); /* and the minimum support */
  if (smin   > 100) error(E_SUPPORT, smin);
  if (bdrcnt < 0)   error(E_NOMEM);
  if ((!fn_inp || !*fn_inp) && (fn_sel && !*fn_sel))
    error(E_STDIN);             /* stdin must not be used twice */
  switch (target) {             /* check and translate target type */
    case 's': target = ISR_ALL;               break;
    case 'c': target = ISR_CLOSED;            break;
    case 'm': target = ISR_MAXIMAL;           break;
    default : error(E_TARGET, (char)target);  break;
  }                             /* (get target type code) */
  switch (covsim) {             /* check and translate sim. measure */
    case 'r': covsim = JIM_RUSSEL_RAO;        break;
    case 'k': covsim = JIM_KULCZYNSKI;        break;
    case 'j': covsim = JIM_JACCARD;           break;
    case 'd': covsim = JIM_DICE;              break;
    case 's': covsim = JIM_SOKAL_SNEATH_1;    break;
    case 'm': covsim = JIM_SOKAL_MICHENER;    break;
    case 'f': covsim = JIM_FAITH;             break;
    case 't': covsim = JIM_ROGERS_TANIMOTO;   break;
    case 'n': covsim = JIM_SOKAL_SNEATH_2;    break;
    case 'o': covsim = JIM_SOKAL_SNEATH_3;    break;
    case 'b': covsim = JIM_BARONI_BUSER;      break;
    case 'g': covsim = JIM_GENERIC;           break;
    default : error(E_SIMILTY, (char)covsim); break;
  }                             /* (get cover similarity measure) */
  switch (eval) {               /* check and translate measure */
    case 'x': eval = JIM_NONE;                break;
    case 'b': eval = JIM_LDRATIO;             break;
    case 'w': eval = JIM_SIZESIM;             break;
    default : error(E_MEASURE, (char)eval);   break;
  }                             /* (get evaluation measure code) */
  if (info == dflt)             /* adapt the default info. format */
    info = (smin < 0) ? " (%a)" : " (%S)";
  mode |= JIM_VERBOSE|JIM_NOCLEAN;
  MSG(stderr, "\n");            /* terminate the startup message */

  /* --- read item selection --- */
  ibase = ib_create(0, 0);      /* create an item base */
  if (!ibase) error(E_NOMEM);   /* to manage the items */
  tread = trd_create();         /* create a transaction reader */
  if (!tread) error(E_NOMEM);   /* and configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  if (fn_sel) {                 /* if item appearances are given */
    CLOCK(t);                   /* start timer, open input file */
    if (trd_open(tread, NULL, fn_sel) != 0)
      error(E_FOPEN, trd_name(tread));
    MSG(stderr, "reading %s ... ", trd_name(tread));
    m = ib_readsel(ibase,tread);/* read the given item selection */
    if (m < 0) error((int)-m, ib_errmsg(ibase, NULL, 0));
    trd_close(tread);           /* close the input file */
    MSG(stderr, "[%"ITEM_FMT" item(s)]", m);
    MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- read transaction database --- */
  tabag = tbg_create(ibase);    /* create a transaction bag */
  if (!tabag) error(E_NOMEM);   /* to store the transactions */
  CLOCK(t);                     /* start timer, open input file */
  if (trd_open(tread, NULL, fn_inp) != 0)
    error(E_FOPEN, trd_name(tread));
  MSG(stderr, "reading %s ... ", trd_name(tread));
  k = tbg_read(tabag, tread, mtar);
  if (k < 0) error(-k, tbg_errmsg(tabag, NULL, 0));
  trd_delete(tread, 1);         /* read the transaction database, */
  tread = NULL;                 /* then delete the table reader */
  m = ib_cnt(ibase);            /* get the number of items, */
  n = tbg_cnt(tabag);           /* the number of transactions, */
  w = tbg_wgt(tabag);           /* the total transaction weight */
  MSG(stderr, "[%"ITEM_FMT" item(s), %"TID_FMT, m, n);
  if (w != (SUPP)n) { MSG(stderr, "/%"SUPP_FMT, w); }
  MSG(stderr, " transaction(s)] done [%.2fs].", SEC_SINCE(t));
  if ((m <= 0) || (n <= 0))     /* check for at least one item */
    error(E_NOITEMS);           /* and at least one transaction */
  MSG(stderr, "\n");            /* terminate the log message */

  /* --- find frequent item sets --- */
  jim = jim_create(target, smin, zmin, zmax, eval, thresh,
                   covsim, simps, sim, algo, mode);
  if (!jim) error(E_NOMEM);     /* create a JIM miner */
  k = jim_data(jim, tabag, sort);
  if (k) error(k);              /* prepare data for JIM */
  report = isr_create(ibase);   /* create an item set reporter */
  if (!report) error(E_NOMEM);  /* and configure it */
  k = jim_report(jim, report);  /* prepare reporter for JIM */
  if (k) error(k);
  if (setbdr(report, w, zmin, &border, bdrcnt) != 0)
    error(E_NOMEM);             /* set the support border */
  if (fn_psp && (isr_addpsp(report, NULL) < 0))
    error(E_NOMEM);             /* set a pattern spectrum if req. */
  if (isr_setfmt(report, scan, hdr, sep, NULL, info) != 0)
    error(E_NOMEM);             /* set the output format strings */
  k = isr_open(report, NULL, fn_out);
  if (k) error(k, isr_name(report));
  if (isr_setup(report) < 0)    /* open the item set file and */
    error(E_NOMEM);             /* set up the item set reporter */
  k = jim_mine(jim);            /* find frequent item sets */
  if (k) error(k);              /* with the JIM algorithm */
  if (stats)                    /* print item set statistics */
    isr_prstats(report, stdout, 0);
  if (isr_close(report) != 0)   /* close item set output file */
    error(E_FWRITE, isr_name(report));

  /* --- write pattern spectrum --- */
  if (fn_psp) {                 /* if to write a pattern spectrum */
    CLOCK(t);                   /* start timer, create table write */
    psp    = isr_getpsp(report);/* get the pattern spectrum */
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* open the output file */
    if (twr_open(twrite, NULL, fn_psp) != 0)
      error(E_FOPEN,  twr_name(twrite));
    MSG(stderr, "writing %s ... ", twr_name(twrite));
    if (psp_report(psp, twrite, 1.0) != 0)
      error(E_FWRITE, twr_name(twrite));
    twr_delete(twrite, 1);      /* write the pattern spectrum */
    twrite = NULL;              /* and delete the table writer */
    MSG(stderr, "[%"SIZE_FMT" signature(s)]", psp_sigcnt(psp));
    MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* write a log message */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */

#endif
